@@ -3,6 +3,7 @@
 # This file is part of IDUtils
 # Copyright (C) 2015-2022 CERN.
 # Copyright (C) 2015-2018 Alan Rubin.
+# Copyright (C) 2025 Will Riley.
 #
 # IDUtils is free software; you can redistribute it and/or modify
 # it under the terms of the Revised BSD License; see LICENSE file for
@@ -14,7 +15,7 @@
 
 """Persistent identifier utilities tests."""
 
-from __future__ import absolute_import, print_function
+import pytest
 
 import idutils
 
@@ -86,6 +87,7 @@ identifiers = [
     ),
     ("ark:/13030/tqb3kh97gh8w", ["ark"], "", ""),
     ("ark:/c8131/g3js3v", ["ark"], "", ""),
+    ("ark:c8131/g3js3v.1", ["ark"], "", ""),
     ("http://www.example.org/ark:/13030/tqb3kh97gh8w", ["ark", "url"], "", ""),
     (
         "10.1016/j.epsl.2011.11.037",
@@ -157,6 +159,7 @@ identifiers = [
         ),
     ),
     ("9783468111242", ["isbn", "ean13"], "978-3-468-11124-2", ""),
+    ("9798847781275", ["isbn", "ean13"], "979-8-8477-8127-5", ""),
     ("978-65-87773-12-4", ["isbn"], "", ""),
     ("4006381333931", ["ean13"], "", ""),
     ("73513537", ["ean8"], "", ""),
@@ -205,19 +208,19 @@ identifiers = [
     ("urn:lsid:ubio.org:namebank:11815", ["lsid", "urn"], "", ""),
     ("0A9 2002 12B4A105 7", ["istc"], "", ""),
     ("1188-1534", ["issn"], "1188-1534", ""),
-    ("12082125", ["pmid"], "12082125", "http://pubmed.ncbi.nlm.nih.gov/12082125"),
-    ("pmid:12082125", ["pmid"], "12082125", "http://pubmed.ncbi.nlm.nih.gov/12082125"),
+    ("12082125", ["pmid"], "12082125", "http://pubmed.ncbi.nlm.nih.gov/12082125/"),
+    ("pmid:12082125", ["pmid"], "12082125", "http://pubmed.ncbi.nlm.nih.gov/12082125/"),
     (
         "https://pubmed.ncbi.nlm.nih.gov/12082125",
         ["pmid", "url"],
         "12082125",
-        "http://pubmed.ncbi.nlm.nih.gov/12082125",
+        "http://pubmed.ncbi.nlm.nih.gov/12082125/",
     ),
     (
         "https://pubmed.ncbi.nlm.nih.gov/12082125/",
         ["pmid", "url"],
         "12082125",
-        "http://pubmed.ncbi.nlm.nih.gov/12082125",
+        "http://pubmed.ncbi.nlm.nih.gov/12082125/",
     ),
     (
         "http://purl.oclc.org/foo/bar",
@@ -243,7 +246,7 @@ identifiers = [
         "",
         "http://nbn-resolving.org/urn:nbn:de:101:1-201102033592",
     ),
-    ("PMC2631623", ["pmcid"], "", "http://www.ncbi.nlm.nih.gov/pmc/PMC2631623"),
+    ("PMC2631623", ["pmcid"], "", "http://pmc.ncbi.nlm.nih.gov/articles/PMC2631623/"),
     (
         "2011ApJS..192...18K",
         ["ads"],
@@ -268,6 +271,12 @@ identifiers = [
         "2017zndo....495787v",
         "http://ui.adsabs.harvard.edu/#abs/2017zndo....495787v",
     ),
+    (
+        "1992ApJ…400L…1W",
+        ["ads"],
+        "1992ApJ...400L...1W",
+        "http://ui.adsabs.harvard.edu/#abs/1992ApJ...400L...1W",
+    ),
     (
         "0000000218250097",
         ["orcid", "isni"],
@@ -477,7 +486,7 @@ identifiers = [
             "refseq",
         ],
         "",
-        "http://www.ncbi.nlm.nih.gov/entrez/viewer.fcgi?val=" "NZ_JXSL01000036.1",
+        "http://www.ncbi.nlm.nih.gov/entrez/viewer.fcgi?val=NZ_JXSL01000036.1",
     ),
     (
         "NM_206454",
@@ -771,11 +780,30 @@ identifiers = [
             ";origin=https://github.com/python/cpython"
         ),
     ),
+    (
+        "custom_scheme_valid",
+        ["custom_scheme"],
+        "custom_scheme_valid",
+        "http://custom/scheme/custom_scheme_valid",
+    ),
+    (
+        "user1@example.com",
+        ["email"],
+        "user1@example.com",
+        "",
+    ),
+    (
+        "5b5056cbf89049064b44d5101764aaae42eb1447",
+        ["sha1"],
+        "5b5056cbf89049064b44d5101764aaae42eb1447",
+        "",
+    ),
 ]
 
 
-def test_detect_schemes():
+def test_detect_schemes(entry_points):
     """Test scheme detection."""
+
     for i, expected_schemes, normalized_value, url_value in identifiers:
         schemes = idutils.detect_identifier_schemes(i)
         assert schemes == expected_schemes, i
@@ -785,7 +813,8 @@ def test_is_type():
     """Test type detection."""
     for i, schemes, normalized_value, url_value in identifiers:
         for s in schemes:
-            assert getattr(idutils, "is_%s" % s)(i)
+            if not s.startswith("custom"):
+                assert getattr(idutils, "is_%s" % s)(i)
 
 
 def test_normalize_pid():
@@ -796,24 +825,19 @@ def test_normalize_pid():
     assert idutils.normalize_pid(None, "handle") is None
 
 
-def test_idempotence():
+def test_idempotence(entry_points):
     """Test persistent id normalization."""
     for i, expected_schemes, normalized_value, url_value in identifiers:
         val_norm = idutils.normalize_pid(i, expected_schemes[0])
         assert expected_schemes[0] in idutils.detect_identifier_schemes(val_norm)
 
 
-def test_to_url():
+def test_to_url(entry_points):
     """Test URL generation."""
     for i, expected_schemes, normalized_value, url_value in identifiers:
-        assert (
-            idutils.to_url(
-                idutils.normalize_pid(i, expected_schemes[0]), expected_schemes[0]
-            )
-            == url_value
-        )
+        assert idutils.to_url(i, expected_schemes[0]) == url_value
         assert idutils.to_url(
-            idutils.normalize_pid(i, expected_schemes[0]),
+            i,
             expected_schemes[0],
             url_scheme="https",
         ) == (
@@ -824,7 +848,7 @@ def test_to_url():
         )
 
 
-def test_valueerror():
+def test_valueerror(entry_points):
     """Test for bad validators."""
     # Many validators rely on a special length of the identifier before
     # testing further. This test, checks that the validators are still
@@ -861,8 +885,82 @@ def test_doi():
     assert not idutils.is_doi("10.1.NOTGOOD.0/123456")
 
 
+def test_url():
+    """Test URL validation."""
+    for i, expected_schemes, normalized_value, url_value in identifiers:
+        if url_value:
+            assert idutils.is_url(url_value)
+
+
 def test_ascl():
     """Test ASCL validation."""
     assert idutils.is_ascl("ascl:1908.011")
     assert idutils.is_ascl("ascl:1908.0113")
     assert not idutils.is_ascl("1990.0803")
+
+
+def test_swh():
+    """Test SWHID validation"""
+
+    swh_with_bad_path_qualifier_value = (
+        "swh:1:cnt:78e48f800c950530e36d3712d9e2e89673f23562"
+        ";origin=https://github.com/python/cpython"
+        ";visit=swh:1:snp:cd510e99a42139ed36f15a5774301c113c3e494b"
+        ";anchor=swh:1:rel:ae1f6af15f3e4110616801e235873e47fd7d1977"
+        ";path=Programs/python.c;lines=12-16"
+    )
+    assert not idutils.is_swh(swh_with_bad_path_qualifier_value)
+
+    swh_with_bad_origin_qualifier_value = (
+        "swh:1:cnt:78e48f800c950530e36d3712d9e2e89673f23562"
+        ";origin=https://github.com/py%thon/cpython"
+        ";visit=swh:1:snp:cd510e99a42139ed36f15a5774301c113c3e494b"
+        ";anchor=swh:1:rel:ae1f6af15f3e4110616801e235873e47fd7d1977"
+        ";path=/Programs/python.c;lines=12-16"
+    )
+    assert not idutils.is_swh(swh_with_bad_origin_qualifier_value)
+
+    swh_with_bad_visit_qualifier_value = (
+        "swh:1:cnt:78e48f800c950530e36d3712d9e2e89673f23562"
+        ";origin=https://github.com/python/cpython"
+        ";visit=https://github.com/python/cpython"
+        ";anchor=swh:1:rel:ae1f6af15f3e4110616801e235873e47fd7d1977"
+        ";path=/Programs/python.c;lines=1-2"
+    )
+    assert not idutils.is_swh(swh_with_bad_visit_qualifier_value)
+
+    swh_with_bad_anchor_qualifier_value = (
+        "swh:1:cnt:78e48f800c950530e36d3712d9e2e89673f23562"
+        ";origin=https://github.com/python/cpython"
+        ";visit=swh:1:snp:cd510e99a42139ed36f15a5774301c113c3e494b"
+        ";anchor=https://github.com/python/cpython"
+        ";path=/Programs/python.c;lines=1-2"
+    )
+    assert not idutils.is_swh(swh_with_bad_anchor_qualifier_value)
+
+    swh_with_bad_lines_qualifier_value = (
+        "swh:1:cnt:78e48f800c950530e36d3712d9e2e89673f23562"
+        ";origin=https://github.com/python/cpython"
+        ";visit=swh:1:snp:cd510e99a42139ed36f15a5774301c113c3e494b"
+        ";anchor=swh:1:rel:ae1f6af15f3e4110616801e235873e47fd7d1977"
+        ";path=/Programs/python.c;lines=12--16"
+    )
+    assert not idutils.is_swh(swh_with_bad_lines_qualifier_value)
+
+    swh_with_bad_lines_qualifier_value = (
+        "swh:1:cnt:78e48f800c950530e36d3712d9e2e89673f23562"
+        ";origin=https://github.com/python/cpython"
+        ";visit=swh:1:snp:cd510e99a42139ed36f15a5774301c113c3e494b"
+        ";anchor=swh:1:rel:ae1f6af15f3e4110616801e235873e47fd7d1977"
+        ";path=/Programs/python.c;lines="
+    )
+    assert not idutils.is_swh(swh_with_bad_lines_qualifier_value)
+
+    swh_with_bad_scheme_version_value = (
+        "swh:2:cnt:78e48f800c950530e36d3712d9e2e89673f23562"
+        ";origin=https://github.com/python/cpython"
+        ";visit=swh:1:snp:cd510e99a42139ed36f15a5774301c113c3e494b"
+        ";anchor=swh:1:rel:ae1f6af15f3e4110616801e235873e47fd7d1977"
+        ";path=/Programs/python.c;lines=1-2"
+    )
+    assert not idutils.is_swh(swh_with_bad_scheme_version_value)
